{-# OPTIONS --type-in-type #-}

open import NeilPrelude
open import Extensionality

module Data.Maybe where

-----------------------------------------------

maybe : {A B : Set} → B → (A → B) → Maybe A → B
maybe b _ nothing = b
maybe _ f (just a) = f a

fromMaybe : {A : Set} → A → Maybe A → A
fromMaybe a nothing  = a
fromMaybe a (just b) = b

maybeMerge : {A B C : Set} → (A → C) → (B → C) → (A → B → C) → Maybe A → Maybe B → Maybe C
maybeMerge fa fb fab nothing nothing   = nothing
maybeMerge fa fb fab nothing (just b)  = just (fb b)
maybeMerge fa fb fab (just a) nothing  = just (fa a)
maybeMerge fa fb fab (just a) (just b) = just (fab a b)

maybeMergeWith : {A : Set} → (A → A → A) → Maybe A → Maybe A → Maybe A
maybeMergeWith f (just a) (just b) = just (f a b)
maybeMergeWith f nothing mb = mb
maybeMergeWith f ma nothing = ma

maybeMap : {A B : Set} → (A → B) → Maybe A → Maybe B
maybeMap f = maybe nothing (just ∘ f)

maybeAp : {A B : Set} → Maybe (A → B) → Maybe A → Maybe B
maybeAp (just f) (just a) = just (f a)
maybeAp _        _        = nothing

maybeJoin : {A : Set} → Maybe (Maybe A) → Maybe A
maybeJoin nothing   = nothing
maybeJoin (just ma) = ma

maybeFilter : {A : Set} → (A → Bool) → Maybe A → Maybe A
maybeFilter p nothing  = nothing
maybeFilter p (just a) with p a
... | true  = just a
... | false = nothing

maybePlus : {A : Set} → Maybe A → Maybe A → Maybe A
maybePlus nothing ma = ma
maybePlus (just a) _ = (just a)

maybeTagL : {A B : Set} → A → Maybe B → Maybe (A × B)
maybeTagL a = maybeMap (_,_ a)

maybeTagR : {A B : Set} → A → Maybe B → Maybe (B × A)
maybeTagR a = maybeMap (λ b → (b , a))

open import Properties

just-inj : {A : Set} → Injective {A} just
just-inj refl = refl

-----------------------------------------------

open import Class.Eq

eqmaybe : {A : Set} → {{ eqA : Eq A }} → Maybe A → Maybe A → Bool
eqmaybe nothing nothing = true
eqmaybe (just a) (just b) = a == b
eqmaybe _ _ = false

eqMaybe : {A : Set} → {{ eqA : Eq A}} → Eq (Maybe A)
eqMaybe = record { _==_ = eqmaybe }

-----------------------------------------------

lem-maybe-fmap-id : {A : Set} → (ma : Maybe A) → maybeMap {A} id ma ≡ ma
lem-maybe-fmap-id nothing = refl
lem-maybe-fmap-id (just _) = refl

lem-maybe-fmap-comp : {A B C : Set} → {f : A → B} → {g : B → C} → (ma : Maybe A) → maybeMap (g ∘ f) ma ≡ (maybeMap g ∘ maybeMap f) ma
lem-maybe-fmap-comp nothing = refl
lem-maybe-fmap-comp (just _) = refl

lem-maybe-fmapPureAp : {A B : Set} → {g : A → B} → {fa : Maybe A} → maybeMap g fa ≡ maybeAp (just g) fa
lem-maybe-fmapPureAp {_} {_} {_} {nothing} = refl
lem-maybe-fmapPureAp {_} {_} {_} {just _}  = refl

lem-maybe-apId : {A : Set} → {fa : Maybe A} → maybeAp (just id) fa ≡ fa
lem-maybe-apId {_} {nothing}  = refl
lem-maybe-apId {_} {just _} = refl

lem-maybe-apCompose : {A B C : Set} → {fa : Maybe A} → {fbc : Maybe (B → C)} → {fab : Maybe (A → B)} →
             maybeAp (maybeAp (maybeAp (just (_∘_)) fbc) fab) fa ≡ maybeAp fbc (maybeAp fab fa)
lem-maybe-apCompose {_} {_} {_} {nothing} {nothing} = refl
lem-maybe-apCompose {_} {_} {_} {nothing} {just _} {nothing} = refl
lem-maybe-apCompose {_} {_} {_} {nothing} {just _} {just _} = refl
lem-maybe-apCompose {_} {_} {_} {just _}  {nothing} = refl
lem-maybe-apCompose {_} {_} {_} {just _}  {just _} {nothing} = refl
lem-maybe-apCompose {_} {_} {_} {just _}  {just _} {just _} = refl

lem-maybe-interchangeLaw  : {A B : Set} → {fab : Maybe (A → B)} → {a : A} → maybeAp fab (just a) ≡ maybeAp (just (λ g → g a)) fab
lem-maybe-interchangeLaw {_} {_} {nothing} = refl
lem-maybe-interchangeLaw {_} {_} {just _}  = refl

lem-maybe-join-assoc : {A : Set} → (mmma : Maybe (Maybe (Maybe A))) → maybeJoin (maybeJoin mmma) ≡ maybeJoin (maybeMap maybeJoin mmma)
lem-maybe-join-assoc nothing = refl
lem-maybe-join-assoc (just _) = refl

lem-maybe-join-unitL  : {A : Set} → (ma : Maybe A) → maybeJoin (just ma) ≡ ma
lem-maybe-join-unitL nothing = refl
lem-maybe-join-unitL (just _) = refl

lem-maybe-join-unitR  : {A : Set} → (ma : Maybe A) → maybeJoin (maybeMap just ma) ≡ ma
lem-maybe-join-unitR nothing = refl
lem-maybe-join-unitR (just _) = refl

lem-maybe-empty-unitR : {A : Set} → {fa : Maybe A} → maybePlus fa nothing ≡ fa
lem-maybe-empty-unitR {_} {nothing} = refl
lem-maybe-empty-unitR {_} {just _} = refl

lem-maybe-mplus-assoc : {A : Set} → {fa fb fc : Maybe A} → maybePlus fa (maybePlus fb fc) ≡ maybePlus (maybePlus fa fb) fc
lem-maybe-mplus-assoc {_} {nothing} {nothing} = refl
lem-maybe-mplus-assoc {_} {nothing} {just _} = refl
lem-maybe-mplus-assoc {_} {just _} = refl

lem-maybe-empty-zeroR : {A B : Set} → {ma : Maybe A} → maybeJoin {B} (maybe nothing (const (just nothing)) ma) ≡ nothing
lem-maybe-empty-zeroR {_} {_} {nothing} = refl
lem-maybe-empty-zeroR {_} {_} {just _} = refl

lem-maybe-join-fmap : {A B : Set} → {f : A → B} → (mma : Maybe (Maybe A)) → maybeJoin (maybeMap (maybeMap f) mma) ≡ maybeMap f (maybeJoin mma)
lem-maybe-join-fmap nothing = refl
lem-maybe-join-fmap (just _) = refl

open import Class.Functor

pointedMaybe : Pointed Maybe
pointedMaybe = record { pure = just }

functorMaybe : Functor Maybe
functorMaybe = record { fmap = maybeMap ;
                        fmap-id = ext lem-maybe-fmap-id  ;
                        fmap-comp = ext lem-maybe-fmap-comp }

pointedfunctorMaybe : PointedFunctor Maybe
pointedfunctorMaybe = record { pointedF = pointedMaybe ; functorF = functorMaybe ; pointedLaw = refl }

applicativeMaybe : Applicative Maybe
applicativeMaybe = record { pointedfunctorF = pointedfunctorMaybe ;
                            _<*>_ = maybeAp ;
                            fmapPureAp = lem-maybe-fmapPureAp ;
                            apId = lem-maybe-apId ;
                            apCompose = lem-maybe-apCompose ;
                            homomorphismLaw = refl ;
                            interchangeLaw = lem-maybe-interchangeLaw }

alternativeMaybe : Alternative Maybe
alternativeMaybe = record { applicativeF = applicativeMaybe ;
                            empty = nothing ;
                            _<|>_ = maybePlus ;
                            empty-unitL = refl ;
                            empty-unitR = lem-maybe-empty-unitR ;
                            mplus-assoc = λ {A} {fa} → lem-maybe-mplus-assoc {A} {fa} }

monadMaybe : Monad Maybe
monadMaybe = record { applicativeM = applicativeMaybe ;
                      join       = maybeJoin ;
                      join-assoc = ext lem-maybe-join-assoc ;
                      join-unitL = ext lem-maybe-join-unitL ;
                      join-unitR = ext lem-maybe-join-unitR ;
                      join-fmap  = ext lem-maybe-join-fmap }

monadPlusMaybe : MonadPlus Maybe
monadPlusMaybe = record { alternativeM = alternativeMaybe ;
                          monadM = monadMaybe ;
                          empty-zeroL = refl ;
                          empty-zeroR = λ {A} {B} {ma} → lem-maybe-empty-zeroR {A} {B} {ma} }

-----------------------------------------------
