{-# OPTIONS --type-in-type #-}

open import NeilPrelude
open import Properties
open import Logic
open import Recursion
open import Data.List
open import Data.Maybe using (just-inj)
open import Data.Nat

module ListProps where

-- Properties of Lists on a Specific Set --

module ListPropsAux {A : Set} where

  ∷-inj : Injective2 {A} _∷_
  ∷-inj refl = refl , refl

  ∷-cong : Congruent2 {A} _∷_
  ∷-cong = cong2 _∷_

  ∷-congL : Congruent2L {A} _∷_
  ∷-congL = cong2L _∷_

  ∷-congR : Congruent2R {A} _∷_
  ∷-congR = cong2R _∷_

  ++assoc : {bs cs : List A} → (as : List A) → (as ++ bs) ++ cs ≡ as ++ bs ++ cs
  ++assoc = listrec refl ∷-congL

  ++[]' : (as : List A) → as ++ [] ≡ as
  ++[]' = listrec refl ∷-congL

  ++[] : {as : List A} → as ++ [] ≡ as
  ++[] {as} = ++[]' as

  import Monoid
  open Monoid {List A} _++_ (λ {as} → ++assoc as) [] refl ++[] public

  import SemiGroup
  open SemiGroup _++_ (λ {as} → ++assoc as) public

  ∷-ninjL : {a b : A} → {as bs : List A} → a ≢ b → a ∷ as ≢ b ∷ bs
  ∷-ninjL p q = p (fst (∷-inj q))

  ∷-ninjR : {a b : A} → {as bs : List A} → as ≢ bs → a ∷ as ≢ b ∷ bs
  ∷-ninjR p q = p (snd (∷-inj q))

  as≠aas : {a : A} → {as : List A} → as ≢ (a ∷ as)
  as≠aas {a} {[]} ()
  as≠aas {a} {b ∷ bs} p = ∷-ninjR as≠aas p

  listIdem : {f : A → A} → ({a : A} → f (f a) ≡ f a) → (as : List A) → map f (map f as) ≡ map f as
  listIdem eIdem = listrec refl (∷-cong eIdem)

  ++length : {bs : List A} → (as : List A) → length (as ++ bs) ≡ length as + length bs
  ++length = listrec refl (cong S)

  ++lengthC : {bs : List A} → (as : List A) → length as + length bs ≡ length (as ++ bs)
  ++lengthC = sym ∘' ++length

  +++length : {cs : List A} → (as bs : List A) → length (as ++ bs ++ cs) ≡ length as + length bs + length cs
  +++length as bs =  trans2 (cong length (assocR {as})) (++length (as ++ bs)) (cong2 _+_ (++length as) refl)

  +++lengthC : {cs : List A} → (as bs : List A) → length as + length bs + length cs ≡ length (as ++ bs ++ cs)
  +++lengthC as bs = sym (+++length as bs)


  lengthZero : {as : List A} → length as ≡ O → as ≡ []
  lengthZero {[]} _ = refl
  lengthZero {_ ∷ _} ()

  mapId' : (as : List A) → map id as ≡ as
  mapId' = listrec refl ∷-congL

  mapId : {as : List A} → map id as ≡ as
  mapId {as} = mapId' as

  -- might need something else (other than just listrec) for ++rev, not sure

  ++rev : {as : List A} → (bs : List A) → rev as bs ≡ reverse bs ++ as
  ++rev []       = refl
  ++rev (b ∷ bs) = trans2 (++rev bs) (assocR {reverse bs}) (cong2R _++_ (sym (++rev bs)))

  slowReverse : List A → List A
  slowReverse [] = []
  slowReverse (a ∷ as) = slowReverse as ++ wrap a

  reversesEq : (as : List A) → reverse as ≡ slowReverse as
  reversesEq = listrec' refl (λ _ as hyp → trans (++rev as) (cong2R _++_ hyp))

  ++reverseAux : {a : A} → {bs : List A} → (as : List A) → reverse (a ∷ as ++ bs) ≡ (reverse (as ++ bs)) ++ wrap a
  ++reverseAux {a} {bs} as = trans (reversesEq (a ∷ as ++ bs)) (cong2R _++_ (sym (reversesEq (as ++ bs))))

  ++reverse : {bs : List A} → (as : List A) → reverse (as ++ bs) ≡ reverse bs ++ reverse as
  ++reverse {bs} = listrec' (sym ++[]) (λ _ as hyp → trans3 (++reverseAux as) (cong2R _++_ hyp) (++assoc (reverse bs)) (cong (_++_ (reverse bs)) (sym (++rev as))))

  revrev' : (as : List A) → reverse (reverse as) ≡ as
  revrev' = listrec' refl (λ a as hyp → trans2 (cong reverse (reversesEq (a ∷ as))) (++reverse (slowReverse as)) (cong (_∷_ a) (trans (cong reverse (sym (reversesEq as))) hyp)))

  revrev : {as : List A} → reverse (reverse as) ≡ as
  revrev {as} = revrev' as

  reverseEq : {as bs : List A} → reverse as ≡ reverse bs → as ≡ bs
  reverseEq eq = trans2 (sym revrev) (cong reverse eq) revrev

  ++reverseAuxR : {as bs : List A} → {b : A} → reverse as ++ b ∷ bs ≡ rev (wrap b) as ++ bs
  ++reverseAuxR {as} {bs} = trans (assocR {rev [] as}) (cong (flip _++_ bs) (sym (++rev as)))

  ++cancL : {bs cs : List A} → (as : List A) → as ++ bs ≡ as ++ cs → bs ≡ cs
  ++cancL = listrec id (_⋙_ (snd ∘ ∷-inj))

  ++cancR : {bs cs : List A} → (as : List A) → as ++ cs ≡ bs ++ cs → as ≡ bs
  ++cancR {bs} {cs} as eq = reverseEq (++cancL (reverse cs) (trans2 (sym (++reverse as)) (cong reverse eq) (++reverse bs)))

  import Cancellative
  open Cancellative _++_ (λ {as} → ++cancL as) (λ {as} → ++cancR as) public

  ++∷-assoc : {bs : List A} → {b : A} → (as : List A) → as ++ (b ∷ bs) ≡ (as ++ wrap b) ++ bs
  ++∷-assoc []       = refl
  ++∷-assoc (a ∷ as) = ∷-congL (++∷-assoc as)

  []≠++wrap : {a : A} → (as : List A) → [] ≢ as ++ wrap a
  []≠++wrap [] ()
  []≠++wrap (_ ∷ _) ()

  ++eq[] : {as bs : List A} → as ++ bs ≡ [] → as ≡ [] × bs ≡ []
  ++eq[] {[]} eq = refl , eq
  ++eq[] {_ ∷ _} ()

  concatMapWrap' : (as : List A) → concat (map wrap as) ≡ as
  concatMapWrap' = listrec refl ∷-congL

  concatMapWrap : {as : List A} → concat (map wrap as) ≡ as
  concatMapWrap {as} = concatMapWrap' as

  concat++ : {bs : List (List A)} → (as : List (List A)) → concat (as ++ bs) ≡ concat as ++ concat bs
  concat++ = listrec' refl (λ a _ hyp → trans (cong (_++_ a) hyp) (assocR {a}))


  -- Element of -----------------------------------------------------------

  infix 3 _∈_

  data _∈_ (a : A) : List A → Set where
    hd : {as : List A} → a ∈ a ∷ as
    tl : {b : A} → {as : List A} → a ∈ as → a ∈ b ∷ as

  index : {a : A} → {as : List A} → a ∈ as → ℕ
  index hd = O
  index (tl el) = S (index el)

  data Lookup (as : List A) : ℕ → Set where
    inside  : (a : A) → (el : a ∈ as) → Lookup as (index el)
    outside : (n : ℕ) → Lookup as (length as + n)

  _!_ : (as : List A) → (n : ℕ) → Lookup as n
  [] ! n = outside n
  (a ∷ as) ! O = inside a hd
  (a ∷ as) ! S n with as ! n
  (a ∷ as) ! S .(index el)      | inside b el = inside b (tl el)
  (a ∷ as) ! S .(length as + n) | outside n   = outside n

  -- Sublist --------------------------------------------------------------

  data _⊆_ : List A → List A → Set where
    ⊆-[]  : {bs : List A} → [] ⊆ bs
    ⊆-hd  : {a : A} → {as bs : List A} → as ⊆ bs → (a ∷ as) ⊆ (a ∷ bs)
    ⊆-tl  : {b : A} → {as bs : List A} → as ⊆ bs → as ⊆ (b ∷ bs)


  -- Prefixes -------------------------------------------------

  module ListPrefix where

      data _⊏_ : List A → List A → Set where
        ⊏-[] : {b : A} → {bs : List A}              → []       ⊏ (b ∷ bs)
        ⊏-∷  : {a : A} → {as bs : List A} → as ⊏ bs → (a ∷ as) ⊏ (a ∷ bs)

      infix 3 _⊏_

      _⊏̸_ : Rel (List A)
      as ⊏̸ bs = Not (as ⊏ bs)

      ⊏-irrefl : Irreflexive _⊏_
      ⊏-irrefl {[]} ()
      ⊏-irrefl {a ∷ as} (⊏-∷ p) = ⊏-irrefl p

      ⊏-trans : Transitive _⊏_
      ⊏-trans ⊏-[]    (⊏-∷ q) = ⊏-[]
      ⊏-trans (⊏-∷ p) (⊏-∷ q) = ⊏-∷ (⊏-trans p q)

      ⊏-asym : Asymmetric _⊏_
      ⊏-asym ⊏-[] ()
      ⊏-asym (⊏-∷ p) (⊏-∷ q) = ⊏-asym p q

      ⊏-cong : {a : A} → {as bs : List A} → (a ∷ as) ⊏ (a ∷ bs) → as ⊏ bs
      ⊏-cong (⊏-∷ p) = p

      ⊏̸-injL :  {a b : A} → {as bs : List A} → a ≢ b → (a ∷ as) ⊏̸ (b ∷ bs)
      ⊏̸-injL p (⊏-∷ q) = p refl

      ⊏̸-injR :  {a b : A} → {as bs : List A} → as ⊏̸ bs → (a ∷ as) ⊏̸ (b ∷ bs)
      ⊏̸-injR p (⊏-∷ q) = p q

      import StrictPartialOrder
      open StrictPartialOrder _⊏_ ⊏-irrefl ⊏-asym ⊏-trans

      _⊑_ : Rel (List A)
      _⊑_ = _≤_

      ⊑-antisym : Antisymmetric _⊑_
      ⊑-antisym = antisymmetric

      data Prefix (as bs : List A) : Set where
        equal    : as ≡ bs → Prefix as bs
        prefixL  : as ⊏ bs → Prefix as bs
        prefixR  : bs ⊏ as → Prefix as bs
        noprefix : as ≢ bs → as ⊏̸ bs → bs ⊏̸ as → Prefix as bs

      comparePrefix : Comparer A → (as bs : List A) → Prefix as bs
      comparePrefix c [] [] = equal refl
      comparePrefix c [] (b ∷ bs) = prefixL ⊏-[]
      comparePrefix c (a ∷ as) [] = prefixR ⊏-[]
      comparePrefix c (a ∷ as) (b ∷ bs) with c a b
      comparePrefix c (a ∷ as) (b ∷ bs) | neq p = noprefix (∷-ninjL p) (⊏̸-injL p) (⊏̸-injL (p ∘ sym))
      comparePrefix c (a ∷ as) (.a ∷ bs) | refl with comparePrefix c as bs
      comparePrefix c (a ∷ as) (.a ∷ .as) | refl | equal refl = equal refl
      comparePrefix c (a ∷ as) (.a ∷ bs) | refl | prefixL p = prefixL (⊏-∷ p)
      comparePrefix c (a ∷ as) (.a ∷ bs) | refl | prefixR p = prefixR (⊏-∷ p)
      comparePrefix c (a ∷ as) (.a ∷ bs) | refl | noprefix n p q = noprefix (∷-ninjR n) (⊏̸-injR p) (⊏̸-injR q)

  -- data Prefix (ls rs  : List A) : Set where
  --   prefixB : ls ≡ rs                       → Prefix ls rs
  --   prefixL : {xs : List A} → ls ++ xs ≡ rs → Prefix ls rs
  --   prefixR : {xs : List A} → rs ++ xs ≡ ls → Prefix ls rs

  -- addToPrefix : {ls rs : List A} → (a : A) → Prefix ls rs → Prefix (a ∷ ls) (a ∷ rs)
  -- addToPrefix a (prefixB refl) = prefixB refl
  -- addToPrefix a (prefixL refl) = prefixL refl
  -- addToPrefix a (prefixR refl) = prefixR refl

  -- comparePrefix : {xs ys : List A} → (ls rs : List A) → ls ++ xs ≡ rs ++ ys → Prefix ls rs
  -- comparePrefix [] [] eq = prefixB refl
  -- comparePrefix [] rs eq = prefixL refl
  -- comparePrefix ls [] eq = prefixR refl
  -- comparePrefix (a ∷ ls) (_  ∷ rs) eq with ∷-inj eq
  -- comparePrefix (a ∷ ls) (.a ∷ rs) eq | refl , eqls = addToPrefix a (comparePrefix ls rs eqls)

  -- Suffixes -------------------------------------------------------------

  module ListSuffix where

      data _⊐_ : List A → List A → Set where
        ⊐-eq : {a : A} →  {as bs : List A} → as ≡ bs → (a ∷ as) ⊐ bs
        ⊐-∷  : {a : A} →  {as bs : List A} → as ⊐ bs → (a ∷ as) ⊐ bs

      infix 3 _⊐_

      ⊐-[] : {a : A} → {as : List A} → (a ∷ as) ⊐ []
      ⊐-[] {a} {[]} = ⊐-eq refl
      ⊐-[] {a} {as ∷ as'} = ⊐-∷ ⊐-[]

      ⊐-∷-sub : {a : A} → {as bs : List A} → bs ⊐ (a ∷ as) → bs ⊐ as
      ⊐-∷-sub (⊐-eq refl) = ⊐-∷ (⊐-eq refl)
      ⊐-∷-sub (⊐-∷ p) = ⊐-∷ (⊐-∷-sub p)

      _⊐̸_ : Rel (List A)
      as ⊐̸ bs = Not (as ⊐ bs)

      ⊐̸-++ : {as : List A} → (bs : List A) → as ⊐̸ (bs ++ as)
      ⊐̸-++ {[]} bs ()
      ⊐̸-++ {a ∷ as} bs p rewrite ++∷-assoc {as} {a} bs with p
      ... | ⊐-∷ q  = ⊐̸-++ (bs ++ wrap a) q
      ... | ⊐-eq q = []≠++wrap bs (++cancR [] q)

      ⊐̸-∷ : {as : List A} → {a : A} → as ⊐̸ (a ∷ as)
      ⊐̸-∷ p = ⊐̸-++ [] (⊐-∷ p)

      ⊐-irrefl : Irreflexive _⊐_
      ⊐-irrefl = ⊐̸-++ []

      ⊐-trans : Transitive _⊐_
      ⊐-trans (⊐-eq refl) q = ⊐-∷ q
      ⊐-trans (⊐-∷ p) q = ⊐-∷ (⊐-trans p q)

      ⊐-∷-inj : {as bs : List A} → {b : A} → as ⊐ (b ∷ bs) → as ⊐ bs
      ⊐-∷-inj (⊐-eq refl) = ⊐-∷ (⊐-eq refl)
      ⊐-∷-inj (⊐-∷ p) = ⊐-∷ (⊐-∷-inj p)

      ⊐-asym : Asymmetric _⊐_
      ⊐-asym (⊐-eq refl) q = ⊐̸-∷ q
      ⊐-asym (⊐-∷ p) q = ⊐-asym p (⊐-∷-inj q)

      ⊐̸-∷-cong : {a b : A} → {as bs : List A} → as ⊐̸ bs → (a ∷ as) ⊐̸ (b ∷ bs)
      ⊐̸-∷-cong p (⊐-eq refl) = p (⊐-eq refl)
      ⊐̸-∷-cong p (⊐-∷ (⊐-eq {a} q)) = ⊐̸-∷-cong {a} (λ x → p (⊐-∷ x)) (⊐-eq q)
      ⊐̸-∷-cong p (⊐-∷ (⊐-∷ {a} q)) = ⊐̸-∷-cong {a} (λ x → p (⊐-∷ x)) (⊐-∷ q)

      ⊐̸-∷-irrefl : {a b : A} → {as : List A} → (a ∷ as) ⊐̸ (b ∷ as)
      ⊐̸-∷-irrefl = ⊐̸-∷-cong ⊐-irrefl

      ⊐-sym-eq : {a b : A} → {as bs : List A} → (a ∷ as) ⊐ bs → (b ∷ bs) ⊐ as → as ≡ bs
      ⊐-sym-eq (⊐-eq refl) q = refl
      ⊐-sym-eq (⊐-∷ p) (⊐-eq refl) = refl
      ⊐-sym-eq (⊐-∷ p) (⊐-∷ q) = absurd (⊐-asym p q)

      ⊐̸-≢→⊐̸ : {a b : A} → {as bs : List A} → as ⊐̸ (b ∷ bs) → as ≢ (b ∷ bs) → (a ∷ as) ⊐̸ (b ∷ bs)
      ⊐̸-≢→⊐̸ p q (⊐-eq r) = q r
      ⊐̸-≢→⊐̸ p q (⊐-∷ r) = p r

      ⊐̸-tail→≢ : {a b : A} → {as bs : List A} → (b ∷ bs) ⊐̸ as → (a ∷ as) ≢ (b ∷ bs)
      ⊐̸-tail→≢ p refl = p (⊐-eq refl)

      ⊐̸-tail : {a : A} → {as bs : List A} → bs ⊐̸ as → bs ⊐̸ (a ∷ as)
      ⊐̸-tail p (⊐-eq refl) = p (⊐-∷ (⊐-eq refl))
      ⊐̸-tail p (⊐-∷ q) = ⊐̸-tail (λ x → p (⊐-∷ x)) q

      import StrictPartialOrder
      open StrictPartialOrder _⊐_ ⊐-irrefl ⊐-asym ⊐-trans

      _⊒_ : Rel (List A)
      _⊒_ = _≤_

      ⊒-antisym : Antisymmetric _⊒_
      ⊒-antisym = antisymmetric

      data Suffix (as bs : List A) : Set where
        equal    : as ≡ bs → Suffix as bs
        suffixL  : bs ⊐ as → Suffix as bs
        suffixR  : as ⊐ bs → Suffix as bs
        nosuffix : as ≢ bs → as ⊐̸ bs → bs ⊐̸ as → Suffix as bs

      compareSuffix : Comparer A → (as bs : List A) → Suffix as bs
      compareSuffix c [] [] = equal refl
      compareSuffix c [] (b ∷ bs) = suffixL ⊐-[]
      compareSuffix c (a ∷ as) [] = suffixR ⊐-[]
      compareSuffix c (a ∷ as) (b ∷ bs) with compareSuffix c (a ∷ as) bs | compareSuffix c as (b ∷ bs)
      compareSuffix c (a ∷ as) (b ∷ ._) | equal refl | _ = suffixL (⊐-eq refl)
      compareSuffix c (a ∷ ._) (b ∷ bs) | _ | equal refl = suffixR (⊐-eq refl)
      compareSuffix c (a ∷ as) (b ∷ bs) | suffixL p | _ = suffixL (⊐-∷ p)
      compareSuffix c (a ∷ as) (b ∷ bs) | _ | suffixR q = suffixR (⊐-∷ q)
      compareSuffix c (a ∷ as) (b ∷ bs) | suffixR p | suffixL q with ⊐-sym-eq p q | c a b
      compareSuffix c (a ∷ as) (._ ∷ ._) | suffixR p | suffixL q | refl | refl = equal refl
      compareSuffix c (a ∷ as) (b ∷ ._) | suffixR p | suffixL q | refl | neq n = nosuffix (∷-ninjL n) ⊐̸-∷-irrefl ⊐̸-∷-irrefl
      compareSuffix c (a ∷ as) (b ∷ bs) | suffixR p | nosuffix e s t = nosuffix (⊐̸-tail→≢ t) (⊐̸-≢→⊐̸ s e) (⊐̸-tail t)
      compareSuffix c (a ∷ as) (b ∷ bs) | nosuffix e s t | suffixL q = nosuffix (⊐̸-tail→≢ s ∘ sym) (⊐̸-tail s) (⊐̸-≢→⊐̸ t (e ∘ sym))
      compareSuffix c (a ∷ as) (b ∷ bs) | nosuffix e s t | nosuffix e' s' t' = nosuffix (⊐̸-tail→≢ t') (⊐̸-≢→⊐̸ s' e') (⊐̸-≢→⊐̸ t (e ∘ sym))

  open ListPrefix public
  open ListSuffix public

  -- data Suffix (ls rs : List A) : Set where
  --   suffixB : ls ≡ rs                       → Suffix ls rs
  --   suffixL : (xs : List A) → xs ++ ls ≡ rs → Suffix ls rs
  --   suffixR : (xs : List A) → xs ++ rs ≡ ls → Suffix ls rs

  -- reverseAux : {bs cs : List A} → (as : List A) → reverse as ++ bs ≡ reverse cs → reverse bs ++ as ≡ cs
  -- reverseAux {bs} as eq = reverseEq (trans2 (++reverse (reverse bs)) (cong (_++_ (reverse as)) revrev) eq)

  -- compareSuffix : {xs ys : List A} → (ls rs : List A) → xs ++ ls ≡ ys ++ rs → Suffix ls rs
  -- compareSuffix {xs} {ys} ls rs eq with comparePrefix (reverse ls) (reverse rs) (trans2 (sym (++reverse xs)) (cong reverse eq) (++reverse ys))
  -- ... | prefixB eq2 = suffixB (reverseEq eq2)
  -- ... | prefixL {xls} eq2 = suffixL (reverse xls) (reverseAux ls eq2)
  -- ... | prefixR {xrs} eq2 = suffixR (reverse xrs) (reverseAux rs eq2)

---------------------------------------------------------------------------

open ListPropsAux public

---------------------------------------------------------------------------

-- Properties of lists over several sets --

lengthMap : {A B : Set} → {f : A → B} → (as : List A) → length (map f as) ≡ length as
lengthMap = listrec refl (cong S)

map++ : {A B : Set} → {f : A → B} → {as₂ : List A} → (as₁ : List A) → map f (as₁ ++ as₂) ≡ map f as₁ ++ map f as₂
map++ = listrec refl ∷-congL

mapmap : {A B C : Set} → {f : B → C} → {g : A → B} → (as : List A) → map f (map g as) ≡ map (f ∘ g) as
mapmap = listrec refl (∷-cong refl)

--------------------------------------------------------------------------
