{-# OPTIONS --type-in-type #-}

-- This is why the --type-in-type flag is inconsistant

open import NeilPrelude

module RussellParadox where

--------------------------------------------------------------

-- This is a slighty modified version of Daniel Peebles' proof,
-- which in turn was based on Robert Dockins's Coq reorganization of Chad E. Brown's proof.

--------------------------------------------------------------

-- A set X is equivalent to a function (ı → X)
-- A set of sets ⟨Set⟩ is equivalent to a function (A → ⟨Set⟩) for some set A.
data ⟨Set⟩ : Set where
  sets : {A : Set} → (A → ⟨Set⟩) → ⟨Set⟩

-- Given a set of sets represented by (f : A → ⟨Set⟩),
-- then the proof that a set (f a) is an element of f is given by the witness (a : A)
data _∈_ : ⟨Set⟩ → ⟨Set⟩ → Set where
  includes : {A : Set} → (f : A → ⟨Set⟩) → (a : A) → f a ∈ sets f

_∉_ : ⟨Set⟩ → ⟨Set⟩ → Set
A ∉ B = Not (A ∈ B)

-- Given a set X and a set of sets f such that X is an element of f,
-- then there must exist an a such that f a ≡ X
invert : {A : Set} → {X : ⟨Set⟩} → {f : A → ⟨Set⟩} → X ∈ sets f → Σ A (λ a → f a ≡ X)
invert (includes f a) = a , refl

-- The set of all sets that are not elements of themselves, with witness.
RT = Σ[ A ∶ ⟨Set⟩ ] (A ∉ A)

-- The set of all sets that are not elements of themselves, without witness.
russell : ⟨Set⟩
russell = sets {RT} fst

-- We show that russell is not an element of russell by assuming that it is
-- and then proving a contradiction by deconstructing that proof.
no : russell ∉ russell
no e with invert e
... | (._ , nelem) , refl = nelem e

-- We show that russell is an element of russell by providing a proof that it is not an element of russell
yes : russell ∈ russell
yes = includes {RT} fst (sets fst , no) -- includes proj₁ (, no)

paradox : False
paradox = no yes

--------------------------------------------------------------

-- Version without --type-in-type

data ⟨Set⟩₁ : Set₁ where
  sets : {A : Set} → (A → ⟨Set⟩₁) → ⟨Set⟩₁

data _∈₁_ : ⟨Set⟩₁ → ⟨Set⟩₁ → Set₁ where
  includes : {A : Set} → (f : A → ⟨Set⟩₁) → (a : A) → f a ∈₁ sets f

data False₁ : Set₁ where

_∉₁_ : ⟨Set⟩₁ → ⟨Set⟩₁ → Set₁
A ∉₁ B = (A ∈₁ B) → False₁

data Σ₀₁ (A : Set) (B : A → Set₁) : Set₁ where
  _,_ : (a : A) → B a → Σ₀₁ A B

data Σ₁₁ (A : Set₁) (B : A → Set₁) : Set₁ where
  _,_ : (a : A) → B a → Σ₁₁ A B

invert₁ : {A : Set} → {X : ⟨Set⟩₁} → {f : A → ⟨Set⟩₁} → X ∈₁ sets f → Σ₀₁ A (λ a → f a ≡ X)
invert₁ (includes f a) = a , refl

RT₁ = Σ₁₁ ⟨Set⟩₁ (λ A → A ∉₁ A)

-- We cannot construct the russell set.
-- russell₁ : ⟨Set⟩₁
-- russell₁ = {!!} -- sets {RT₁} fst

--------------------------------------------------------------
